import os
import unittest

from hwt.hdlObjects.value import Value
from hwt.simulator.agentConnector import valToInt
from hwt.simulator.hdlSimulator import HdlSimulator
from hwt.simulator.simSignal import SimSignal
from hwt.simulator.vcdHdlSimConfig import VcdHdlSimConfig
from hwt.simulator.configVhdlTestbench import HdlSimConfigVhdlTestbench
from hwt.simulator.utils import agent_randomize

def allValuesToInts(sequenceOrVal):
    if isinstance(sequenceOrVal, Value):
        return valToInt(sequenceOrVal)
    elif not sequenceOrVal:
        return sequenceOrVal
    else:
        l = []
        for i in sequenceOrVal:
            l.append(allValuesToInts(i))
        return l

class SimTestCase(unittest.TestCase):
    """
    This is TestCase class contains methods which are usually used during
    hdl simulation.
    
    @attention: self.model, self.procs has to be specified before running doSim
    u = Axi_rDatapump()
    self.model, self.procs = simPrepare(u)
    
    """
    
    def getTestName(self):
        className, testName = self.id().split(".")[-2:]
        return "%s_%s" % (className, testName)
    
    def doSim(self, time):
        outputFileName = "tmp/" + self.getTestName() + ".vcd"
        d = os.path.dirname(outputFileName)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFileName, 'w') as outputFile:
            # return _simUnitVcd(simModel, stimulFunctions, outputFile=f, time=time) 
            sim = HdlSimulator()

            # configure simulator to log in vcd
            sim.config = VcdHdlSimConfig(outputFile)
            
            # run simulation, stimul processes are register after initial initialization
            sim.simUnit(self.model, time=time, extraProcesses=self.procs) 
            return sim
    
    def dumpHdlTestbench(self, time, file=None):
        if file:
            outputFileName = file
        else:
            outputFileName = "tmp/" + self.getTestName() + "_tb.vhd"
        d = os.path.dirname(outputFileName)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFileName, 'w') as outputFile:
            # return _simUnitVcd(simModel, stimulFunctions, outputFile=f, time=time) 
            sim = HdlSimulator()

            # configure simulator to log in vcd
            sim.config = HdlSimConfigVhdlTestbench(self.u)
            
            # run simulation, stimul processes are register after initial initialization
            sim.simUnit(self.model, time=time, extraProcesses=self.procs)
            
            sim.config.dump(outputFile)
            
            return sim
        
    def assertValEqual(self, first, second, msg=None):
        if isinstance(first, SimSignal):
            first = first._val
        if not isinstance(first, int):    
            first = valToInt(first)
        
        return unittest.TestCase.assertEqual(self, first, second, msg=msg)
    
    def assertValSequenceEqual(self, seq1, seq2, msg=None, seq_type=None):
        """
        @param seq1: can contain instance of values or nested list of them
        @param seq2: items are not converted
        """
        seq1 = allValuesToInts(seq1)
        unittest.TestCase.assertSequenceEqual(self, seq1, seq2, msg, seq_type)
        

    def randomize(self, intf):
        self.procs.append(agent_randomize(intf._ag))



from hwt.hdlObjects.constants import Time
from random import Random

def valueHasChanged(valA, valB):
    return valA.val is not valB.val or valA.vldMask != valB.vldMask

def agent_randomize(agent, timeQuantum=5 * 10 * Time.ns, seed=317):
    random = Random(seed)
    def randomEnProc(simulator):
        while True:
            agent.enable = random.random() < 0.5
            delay = int(random.random() * timeQuantum)  
            yield simulator.wait(delay)
    return randomEnProc


from hwt.simulator.agentBase import SyncAgentBase
from hwt.hdlObjects.constants import NOP


class HandshakedAgent(SyncAgentBase):
    """
    Simulation/verification agent for handshaked interface
    @attention: requires clk and rst/rstn signal 
      (if you do not have any create simulation wrapper with it)
    """
    def __init__(self, intf, clk=None, rstn=None):
        super().__init__(intf, clk=clk, rstn=rstn)
        self.actualData = NOP
        self.data = []
        # these signals are extracted like this to make 
        # agent more configurable
        self._rd = self.getRd()
        self._vld = self.getVld()
    
    def getRd(self):
        """get "ready" signal"""
        return self.intf.rd
    
    def getVld(self):
        """get "valid" signal"""
        return self.intf.vld
        
    def monitor(self, s):
        """
        Collect data from interface
        """
        if s.r(self.rst_n).val and self.enable:
            s.w(1, self._rd)
            
            yield s.updateComplete
            vld = s.r(self._vld)
            if vld.val or not vld.vldMask:
                d = self.doRead(s)
                self.data.append(d)
        else:
            s.w(0, self._rd)
    
    def doRead(self, s):
        """extract data from interface"""
        return s.read(self.intf.data)
        
    def doWrite(self, s, data):
        """write data to interface"""
        s.w(data, self.intf.data)
        
    def driver(self, s):
        """Push data to interface"""
        if self.actualData is NOP and self.data:
            self.actualData = self.data.pop(0)
        
        do = self.actualData is not NOP
        
        if do:
            self.doWrite(s, self.actualData)
        else:
            self.doWrite(s, None)
            
        if s.r(self.rst_n).val and do and self.enable:
            s.w(1, self._vld)
        else:
            s.w(0, self._vld)
            return
        
        yield s.updateComplete
        
        rd = s.r(self._rd) 
        if rd.val or not rd.vldMask:
            if self.data:
                self.actualData = self.data.pop(0)
            else:
                self.actualData = NOP

class HandshakeSyncAgent(HandshakedAgent):
    """
    Simulation/verification agent for HandshakedSycn interface
    @attention: there is no data channel on this interface it is synchronization only
    """  
    
    def doWrite(self, s, data):
        pass
    
    def doRead(self, s):
        raise NotImplementedError()    
    
